//********************************************************************
//*** ResourcePool/oracle/SQLExecute.t.cc
//*** Copyright (c) 2003-2008 by Markus Winand <mws@fatalmind.com>
//*** $Id: SQLExecute.t.cc,v 1.18 2008/01/27 16:42:29 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#include "ResourcePool/oracle/ResourcePool.hh"
#include "t/main.hh"

#include "testconfig.cc"

#include "../SQL/SQLExecute.t.cc"

std::auto_ptr<RP> rp;

#ifndef RESOURCEPOOL_ORACLE_ORACLETYPE_HH
#include "oracleType.hh"
#endif
using fatalmind::oracle::oracleType;
using fatalmind::oracle::oracleTypeImpl;

void test_init()
{
    if (rp.get() == 0) {
        ResourcePoolOptions opt;
        opt.setMaxExecTry(1);
        rp.reset(new RP(getFactory(), opt));
    }
}

void INIT_oracle_SQLExecute() {
    rp.reset(new RP(getFactory()));
    test_SQLExecute_init(*rp);
}

void TEST_oracle_SQLExecute_insert() {
	test_SQLExecute_one(*rp);
}

void TEST_oracle_SQLExecute_insert_two() {
	test_SQLExecute_two(*rp);
}

void TEST_oracle_SQLExecute_insert_three() {
	test_SQLExecute_three(*rp);
}

void TEST_PERF_oracle_SQLExecute_insert() {
	test_perf_SQLExecute_insert(*rp);
}

void TEST_PERF_oracle_SQLExecute_insert_outer() {
	test_perf_SQLExecute_insert_outer(*rp);
}

void TEST_PERF_oracle_SQLExecute_update_outer() {
	test_perf_SQLExecute_update_outer(*rp);
}

void TEST_oracle_SQLExecute_insert_fail() {
	test_SQLExecute_fail(*rp);
}

void init2(RP& p) {
        drop(p);
        RP::Command::SQLExecute sql(
                "create table hirsch (id numeric not null primary key, ukn varchar(10))"
        );
        p.execute(sql);
}

void x_TEST_oracle_SQLExecute_batched() {
	init2(*rp);
	clear(*rp);
	std::string insertsql = "insert into hirsch (id, ukn) values (?, ?)";
	RP::Command::SQLExecute param(*rp, insertsql);
	RP::Command::SQLExecute param2(*rp, insertsql);
	RP::Command::SQLExecute param3(*rp, insertsql);
	RP::Command::SQLExecute cmd(*rp, insertsql);
	cmd.bind(0, 900);
	cmd.bind<std::string>(1, "testme1");

	param.bind(0, 901);
	param.bind<std::string>(1, "testme b");
	param2.bind(0, 902);
	param2.bind<std::string>(1, "testme c");
	param3.bind(0, 903);
	param3.bind<std::string>(1, "testme d");
	
	assertTrue(cmd.addBatch(param));
	assertTrue(cmd.addBatch(param2));
	assertTrue(cmd.addBatch(param3));

	rp->execute(cmd);

	RP::Command::SQLSelectRow sel(*rp, "select id from hirsch where id=?");
	int out = 0;
	sel.bindout(0, out);
	sel.bind(0, 901);
	rp->execute(sel);

	// update it
	assertEqual(out, 901);
	std::string updatesql = "update hirsch set ukn = ? where id <= ?";
	RP::Command::SQLExecute up(*rp, updatesql);
	up.bind<std::string>(0, "updated");
	up.bind(1, 1000); // hits all 4 records

	RP::Command::SQLExecute param4(*rp, updatesql);
	param4.bind<std::string>(0, "fits");
	param4.bind(1, 900); // hits one record

	RP::Command::SQLExecute param5(*rp, updatesql);
	param5.bind<std::string>(0, "also too");
	param5.bind(1, 902);

	assertTrue(up.addBatch(param4));
	assertTrue(up.addBatch(param5));
	rp->execute(up);

	std::string result;
	RP::Command::SQLSelectRow sel2(*rp, "select ukn from hirsch where id = ?");
	sel2.bind(0, 903);
	sel2.bindout(0, result);
	rp->execute(sel2);
	assertEqual(result, "updated");

	// check re-execution of up
	RP::Command::SQLExecute up2(*rp, "update hirsch set ukn = 'changed' where id = 903");
	rp->execute(up2);
	rp->execute(sel2);
	assertEqual(result, "changed");

	rp->execute(up);
	rp->execute(sel2);
	assertEqual(result, "updated");
}

void x_TEST_oracle_SQLExecute_batched_fail() {
	init2(*rp);
	clear(*rp);
	std::string insertsql = "insert into hirsch (id, ukn) values (?, ?)";
	std::string insertsql2 = "insert into HIRSCH    (id, ukn) values (?, ?)";
	RP::Command::SQLExecute param(*rp, insertsql2);
	RP::Command::SQLExecute cmd(*rp, insertsql);
	cmd.bind(0, 900);
	cmd.bind<std::string>(1, "testme1");
	param.bind(0, 901);
	param.bind<std::string>(1, "testme b");

	assertFalse(cmd.addBatch(param));

	RP::Command::SQLExecute param2(*rp, insertsql);
	try {
		cmd.addBatch(param2);
		// number of parameters don't match
		assertNotReach();
	} catch (...) {
	}

	param2.bind(0, 100);	
	param2.bind(1, 101); // different type
	assertFalse(cmd.addBatch(param2));
	assertEqual(cmd.batchSize(), 1); // only one statement
	
	rp->execute(cmd); // execution is possible

	
	RP::Command::SQLSelectRow sel(*rp, "select count(*) from hirsch");
	int count = 0 ;
	sel.bindout(0, count);
	rp->execute(sel);
	assertEqual(count, 1); // only one statement executed

	// TODO, test correct throwing of exceptions in case of SQL errors

	RP::Command::SQLExecute exceptioncmd(*rp, insertsql);
	exceptioncmd.bind(0, 2000);
	exceptioncmd.bind<std::string>(1, "works");


	RP::Command::SQLExecute exceptioncmd2(*rp, insertsql);
	exceptioncmd2.bind(0, 2000);
	//exceptioncmd2.bind<std::string>(1, "failes since it is way too long to fit into the column");
	exceptioncmd2.bind<std::string>(1, "blah");
	assertTrue(exceptioncmd.addBatch(exceptioncmd2));
	// TODO: define TX handling in case of OCI_BATCH_ERRORS and failure!
	RP::Command::Transaction t(exceptioncmd);
	for (int i = 0; i < 500; ++i) {
		assertNotGrowing(0);
		try {
			rp->execute(t);
			assertNotReach();
		} catch (fatalmind::Exception e) {
			std::cout << e.message() << std::endl;
		}
	}
}

void CLEANUP_oracle_SQLExecute() {
	drop(*rp);
}
